#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#define TAM_MAX_FINGER 32

// Estrutura para armazenar um nó na DHT
typedef struct Node {
    int id;
    int *finger; //array para a tabela finger
    int tam_finger;
    int *keys; //array para as chaves armazenadas
    int key_count;
    struct Node* next;
    struct Node* prev;
} Node;


int atualiza_max_id(Node *dht){
    Node *current = dht;
    int max_id = 0;
    do {
        if(current->id > max_id){
            max_id = current->id;
        }
        current = current->next;
    } while (current != dht);
    return max_id;
}

void redistribute_keys(Node *dht, Node* new_node) {
    Node *start = dht;
    do {
        Node* current = new_node;
        // printf("Start:%d Next:%d\n", start->id, current->id);
        // printf("Start->key_count:%d\n", start->key_count);
        //O novo nodo deve ter as chaves EXISTENTES que são maiores que o antecessor e menores ou iguais a ele
        for (int i = 0; i < start->key_count; i++) {
            // printf("Start->keys[%d]:%d\n", i, start->keys[i]);
            if (start->keys[i] > start->id && start->keys[i] <= current->id) {
                current->keys[current->key_count++] = start->keys[i];
                start->keys[i] = 0;
                start->key_count--;
            }
        }
        start = start->next;
    } while (start != new_node);
}

// Função para criar um novo nó
Node* create_node(int id, int max_id) {
    Node* node = (Node*)malloc(sizeof(Node));
    node->id = id;
    node->key_count = 0;
    node->finger = (int*)malloc(TAM_MAX_FINGER* 10 * sizeof(int));
    //O tamanho do array de chave é calculado por id - 1 < k <= id
    node->keys = (int*)malloc((id - 1)*sizeof(int));
    node->next = NULL;
    node->prev = NULL;
    return node;
}

// Função para inserir um nó na DHT (circular e ordenado)
Node *insert_node(Node *head, int id, int max_id) {
    Node* new_node = create_node(id, max_id);
    if (head == NULL) {
        new_node->next = new_node;
        new_node->prev = new_node;
        return new_node;
    }
    Node* current = head;
    Node* prev = NULL;
    
    // Caso especial: inserir antes do nó cabeça
    if (id < head->id) {
        // Encontrar o último nó para atualizar seu ponteiro
        Node* last = head;
        while (last->next != head) {
            last = last->next;
        }
        last->next = new_node;
        new_node->next = head;
        new_node->prev = last;
        head->prev = new_node;
        redistribute_keys(head, new_node);
        return new_node; // Novo nó se torna a nova cabeça
    }

    // Inserir em posição ordenada
    do {
        prev = current;
        current = current->next;
    } while (current != head && current->id < id);

    prev->next = new_node;
    new_node->next = current;
    new_node->prev = prev;
    current->prev = new_node;
    redistribute_keys(head, new_node);

    return head;
}

void imprime_lista(Node* head) {
    Node* current = head;
    do {
        printf("%d -> ", current->id);
        current = current->next;

    } while (current != head);
}

void transfer_keys(Node* from, Node* to) {
    int* new_keys = (int*)malloc((from->key_count + to->key_count) * sizeof(int));
    for (int i = 0; i < to->key_count; i++) {
        new_keys[i] = to->keys[i];
    }
    for (int i = 0; i < from->key_count; i++) {
        new_keys[to->key_count + i] = from->keys[i];
    }
    free(to->keys);
    to->keys = new_keys;
    to->key_count += from->key_count;
}


Node* remove_node(Node* head, int id) {
    if (head == NULL) {
        return NULL; // Lista vazia
    }

    Node* current = head;
    Node* prev = NULL;

    // Caso especial: o nó a ser removido é o nó cabeça
    if (head->id == id) {
        if (head->next == head) {
            // A lista contém apenas um nó
            free(head);
            return NULL;
        } else {
            // Encontrar o último nó para atualizar seu ponteiro
            Node* last = head;
            while (last->next != head) {
                last = last->next;
            }
            last->next = head->next;
            head->next->prev = last;
            Node* new_head = head->next;
            transfer_keys(head, new_head);
            free(head);
            return new_head;
        }
    }

    // Percorrer a lista para encontrar o nó a ser removido
    do {
        prev = current;
        current = current->next;
    } while (current != head && current->id != id);

    if (current->id == id) {
        prev->next = current->next;
        current->next->prev = prev;
        transfer_keys(current, current->next);
        free(current);
    }

    return head;
}


// Função para calcular a tabela finger
void calculate_finger_table(Node* head, int max_id) {
    Node* current = head;
    // printf("Max_id:%d\n", max_id);
    int log2_max_id = ceil(log2(max_id));
    // printf("Log2_max_id:%d\n", log2_max_id);
    do {
        for (int k = 1; k <= log2_max_id; k++) {
            // printf("N:%d", current->id);
            int finger_id = (current->id + (1 << (k-1))) % (1 << log2_max_id);
            // printf(" Finger_id:%d\n", finger_id);
            Node* temp = head;
            //Inserir na tabela_finger o id do nó que é o menor id maior que o id do finger_id, ou seja, o id mais próximo do finger_id
            while(temp->id < finger_id && temp->next != head) {
                temp = temp->next;
            }
            if(temp->next == head && temp->id < finger_id){
                temp = head;
            }
            if(temp->id == current->id)
                temp = temp->next;
            current->finger[k-1] = temp->id;
        }
        current->tam_finger = log2_max_id;
        current = current->next;
    } while (current != head);
}


//Função para imprimir todas as tabelas finger
void print_finger_table(Node** nodes_involved,int num_nodes, int timestamp) {
    for(int i = 0; i < num_nodes; i++) {
        printf("%d T %d {", timestamp, nodes_involved[i]->id);
        for (int j = 0; j < nodes_involved[i]->tam_finger; j++) {
            printf("%d", nodes_involved[i]->finger[j]);
            if (j < nodes_involved[i]->tam_finger - 1)
                printf(",");
        }
        printf("}\n");
    }
}

void print_finger_table_completa(Node* head, int timestamp) {
    Node* current = head;
    do {
        printf("%d F %d {", timestamp, current->id);
        for (int i = 0; i < TAM_MAX_FINGER; i++) {
            printf("%d -> ", current->finger[i]);
        }
        printf("%d}\n", current->id);
        current = current->next;
    } while (current != head);
}



// Função para realizar lookup de uma chave
Node ** lookup(Node* head, int timestamp, int node_id, int key, int *num_nodes, int max_id) {
    // printf("Lookup\n");
    Node* current = head;
    Node** nodes_involved = (Node**)malloc(TAM_MAX_FINGER * sizeof(Node*));
    *num_nodes = 0;

    while (current->id != node_id) {
        current = current->next;
    }

    printf("%d L %d {", timestamp, key);
    Node *start = current;
    int id_aux = current->id;
    nodes_involved[(*num_nodes)++] = current;
    while(1){
        int found = 0;
        printf("%d", current->id);
        //Verificar se o nó atual já tem a chave
        for (int i = 0; i < current->key_count; i++) {
            if (current->keys[i] == key) {
                printf("}\n");
                return nodes_involved;
            }
        }
        printf(",");

        //Procura o maior nó conhecido pela finger_table que está acima do valor da chave
        for (int i = 0; i < current->tam_finger; i++) {
            if (current->finger[i] > key) {
                id_aux = current->finger[i];
                found = 1;
                break;
            }
        }

        //Se não encontrar um nó na tabela finger que seja maior que o valor da chave, procura o maior valor que antecede ela
        if(!found){
            int max = current->finger[0];
            for (int i = 1; i < current->tam_finger; i++) {
                if (current->finger[i] > max) {
                    max = current->finger[i];
                }
            }
            id_aux = max;
        }
        // Mover para o nó com ID id_aux
        while(current->id != id_aux){
            current = current->next;
        }

        //Previne caso onde a chave nao esta no maior nodo 
        if(current->id == max_id && key > current->id){   
            current = head;
        }
        
        nodes_involved[(*num_nodes)++] = current; //Adiciona o nó atual ao array de nós envolvidos

        //Previne loops infinitos
        if(current->id == start->id){
            printf("}\n");
            nodes_involved[(*num_nodes)++] = current;
            break;
        }
    }
    return nodes_involved;
}

// Função para adicionar uma chave a um nó respeitando o limite de um nó guarda as chaves maiores que seu antecessor e menores e iguais a ele.
void insert_key(Node* head, int node_id, int key) {
    Node* current = head; //Lista circular

    //Procuramos o nó com o id igual ao node_id
    while (current->id != node_id) {
        current = current->next;
    }
    
    //A partir no nosso nó, procuramos o próximo nó que cabe a chave
    while(current->id < key && current != head){
        current = current->next;
    }

    current->keys[current->key_count++] = key;
    // printf("Chave %d inserida no nó %d\n", key, current->id);
}

void imprime_chaves(Node* head) {
    Node* current = head;
    do {
        printf("%d -> ", current->id);
        for (int i = 0; i < current->key_count; i++) {
            printf("%d ", current->keys[i]);
        }
        printf("\n");
        current = current->next;
    } while (current != head);
}

// Função principal para testar a inserção de nós
int main() {
    Node* dht = NULL;
    Node **nodes_involved = NULL;
    int timestamp, id, key;
    int max_id = 0;
    char op;

    while (scanf("%d %c %d %d", &timestamp, &op, &id, &key) != EOF) {
        if(id > max_id) {
            max_id = id;
        }
        if (op == 'E') {
            dht = insert_node(dht, id, max_id);
            calculate_finger_table(dht, max_id);
        } else if (op == 'S') {
            dht = remove_node(dht, id);
            max_id = atualiza_max_id(dht);
        } else if (op == 'I') {
            insert_key(dht, id, key);
        } else if (op == 'L') {
            int num_nodes = 0;
            nodes_involved = lookup(dht, timestamp, id, key, &num_nodes, max_id);
            print_finger_table(nodes_involved, num_nodes, timestamp);
            free(nodes_involved);
        } else if (op == 'P'){
            imprime_chaves(dht);
            print_finger_table_completa(dht, timestamp);
        }
    }

    return 0;
}